import numpy as np
import pandas as pd
import scipy as sp
import xgboost

import datagen_modules
import xgboost_modules
import data_generator_1
import data_generator_2

# generate noise
def generate_noise(addedVar, noise_mean, noise_sigma, noiseTF):
    # NOISE
    if noiseTF: 
        noise = np.random.normal(loc=noise_mean, scale=noise_sigma, size=addedVar.shape)
    else:
        noise = 0
    return noise

# Train models for E[Y | v] using trainDataset
def train_Y(X_train,y_train):
    # Compute theta_{k,2} = E[\theta_k1 | Vk, pre(Vk)]
    regmode = xgboost_modules.checkMode(y_train)
    model = xgboost_modules.fit_xgboost(X=X_train, y=y_train, regtype=regmode)
    return model

# Train models for Case 2 -- {P(Vk | pre(Vk) \intersect Vs, V_{\overline{S}}
def train_condProb_case2(X_train, topoSort, S):
    pred_models = {}

    # Sorting S w.r.t topo order
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)

    # For variables in S,
    for idx in range(len(topoSort) - 1):
        # Set Vi
        varidx = topoSort[idx]
        # If i not in S, pass
        if varidx not in S_sorted:
            continue

        Vi = X_train[varidx]

        # Choose pre(Vi) \intersect Vs
        if idx != 0:
            prev_indices = topoSort[:idx]
            intersected = [x for x in S_sorted if x in prev_indices and x in S_sorted]
        else:
            prev_indices = []
            intersected = []

        # Choose V_{\overline{S}}
        Sbar = [x for x in topoSort[:-1] if x not in S_sorted]

        # pre(Vi) \intersect Vs, V_{\overline{S}
        condIDX = Sbar + intersected

        # Choose CondVar =  pre(Vi) \intersect Vs, V_{\overline{S}
        if len(condIDX) != 0:
            CondVar = X_train[condIDX]
            regmode = xgboost_modules.checkMode(Vi)
            model = xgboost_modules.fit_xgboost(X=CondVar, y=Vi, regtype=regmode)
            pred_models[varidx] = model

        else:
            CondVar = []
            # Compute P(Vi)
            ### Take a unique values
            uniqueVi = sorted(pd.unique(Vi))

            # Compute P(vi) for all vi
            prob = np.array([np.sum(Vi == x) / len(Vi) for x in uniqueVi])

            # append pred_models
            pred_models[varidx] = prob

    return pred_models

# Evaluate for {P(Vk | pre(Vk) \intersect Vs, V_{\overline{S}} for all samples in the Test
def estimate_condProb_case2(X_test, allCondModel_S, S, topoSort,noiseTF):
    '''
    Output: allCondProb
    '''

    eval_models = {}

    # Sorting S w.r.t topo order
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)

    # For variables in S,
    for idx in range(len(topoSort)-1):
        # Set Vi
        varidx = topoSort[idx]
        # If i not in S, pass
        if varidx not in S_sorted:
            continue
        # print(idx,varidx)
        Vi = X_test[varidx]

        # Choose pre(Vi) \intersect Vs
        if idx != 0:
            prev_indices = topoSort[:idx]
            intersected = [x for x in S_sorted if x in prev_indices and x in S_sorted]
        else:
            prev_indices = []
            intersected = []

        # Choose V_{\overline{S}}
        Sbar = [x for x in topoSort[:-1] if x not in S_sorted]

        # pre(Vi) \intersect Vs, V_{\overline{S}
        condIDX = Sbar + intersected

        # Call the condmodel P(Vi | condVar)
        model = allCondModel_S[varidx]

        # Choose CondVar =  pre(Vi) \intersect Vs, V_{\overline{S}
        if len(condIDX) == 0:
            CondVar = []
            # Take a unique values
            uniqueVi = sorted(pd.unique(Vi))

            # Compute P(vi) for all vi
            prob_Vi = np.asarray([model[uniqueVi.index(x)] for x in Vi])
            eval_models[varidx] = prob_Vi
        else:
            # Check the cardinality of Vi
            regmode = xgboost_modules.checkMode(Vi)

            # Set CondVar from test dataset
            CondVar = X_test[condIDX]

            # Evalute P(Vi | CondVar)
            prob_Vi = xgboost_modules.predict_xgboost(model, X=CondVar, y=np.asarray(Vi), regtype=regmode)
            eval_models[varidx] = prob_Vi

    # Add noises
    for k in eval_models.keys():
        eval_models[k] += generate_noise(eval_models[k], X_test.shape[0] ** (-1 / 4), 0.1,noiseTF)

    return eval_models


# Train models for {P(Vi | pre(Vi))} using trainDataset
def train_condProb(X_train, topoSort):
    '''
        Output: allCondProb for trained dataset
        Train with P(vi | pre(vi))
    '''
    pred_models = {}

    # For variables in topoSort
    Vi = X_train[topoSort[0]]

    # Take a unique values
    uniqueVi = sorted(pd.unique(Vi))

    # Compute P(vi) for all vi
    prob = np.array([np.sum(Vi == x) / len(Vi) for x in uniqueVi])

    # append pred_models
    pred_models[topoSort[0]] = prob

    for idx in range(1,len(topoSort)-1):
        # Set Vi
        varidx = topoSort[idx]
        Vi = X_train[varidx]

        # pre(Vi)
        pre_Vi = X_train[topoSort[:idx]]
        regmode = xgboost_modules.checkMode(Vi)
        model = xgboost_modules.fit_xgboost(X=pre_Vi, y=Vi, regtype=regmode)
        pred_models[varidx] = model

    return pred_models


# Evaluate for {P(Vi | pre(Vi))} for all samples in the Test
def estimate_condProb(X_test, allCondModel, topoSort, noiseTF):
    '''
    Output: allCondProb
    '''

    eval_models = {}

    ''' For V0
    '''
    # Call the model
    prob = allCondModel[topoSort[0]]

    # For variables in topoSort
    Vi = X_test[topoSort[0]]

    # Take a unique values
    uniqueVi = sorted(pd.unique(Vi))

    # Compute P(vi) for all vi
    prob_Vi = np.asarray([prob[uniqueVi.index(x)] for x in Vi])
    eval_models[topoSort[0]] = prob_Vi

    for idx in range(1, len(topoSort)-1):
        # Set Vi
        varidx = topoSort[idx]
        Vi = X_test[varidx]

        # Call the condmodel P(Vi | pre(Vi))
        model = allCondModel[varidx]

        # Set pre(Vi) from test dataset
        pre_Vi = X_test[topoSort[:idx]]

        # Evalute P(Vi | pre(Vi))
        regmode = xgboost_modules.checkMode(Vi)
        prob_Vi = xgboost_modules.predict_xgboost(model, X=pre_Vi, y=np.asarray(Vi), regtype=regmode)

        # append pred_models
        eval_models[varidx] = prob_Vi

    # Add noises
    for k in eval_models.keys():
        eval_models[k] += generate_noise(eval_models[k], X_test.shape[0] ** (-1 / 4), 0.1,noiseTF)

    return eval_models

# It encodes the subset S as a key string.
# All subset S takes the same key iff they contain the same element.
def keygen(S_sorted):
    return ''.join([str(x) for x in S_sorted])

# Compute E[Y | V], E[Y | v_{S}, V_{\overline{S}}]
def constructTheta_S_case2(X_train, y_train, X_test, y_test, model_Y_train, model_Y_test, x, S, topoSort,noiseTF):
    N = X_test.shape[0]

    # Sort S based on the topological error.
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    S_sorted_reverse = S_sorted.copy()
    S_sorted_reverse.reverse()

    # theta_{1} = Y
    theta_1_train = y_train
    theta_1_test = y_test

    # theta_{2} = E[Y | V]
    regmode = xgboost_modules.checkMode(theta_1_train)

    if regmode == "binary":
        theta_2_test = model_Y_train.predict(xgboost.DMatrix(X_test))
        theta_2_train = model_Y_test.predict(xgboost.DMatrix(X_train))
    else:
        theta_2_test = xgboost_modules.predict_xgboost(model_Y_train, X_test, y_test, regmode)
        theta_2_train = xgboost_modules.predict_xgboost(model_Y_test, X_train, y_train, regmode)

    # theta_{3} = E[Y | v_S, V_S]
    ## Construct v_{S}, V_{S}
    ### Train
    X_train_copy = X_train.copy()
    X_test_copy = X_test.copy()
    for k in S:
        X_train_copy[k] = np.repeat(x[k].values[0],X_train_copy.shape[0])
        X_test_copy[k] = np.repeat(x[k].values[0], X_test_copy.shape[0])

    ### Compute
    if regmode == "binary":
        theta_3_train = model_Y_test.predict(xgboost.DMatrix(X_train_copy))
        theta_3_test = model_Y_train.predict(xgboost.DMatrix(X_test_copy))
    else:
        theta_3_train = xgboost_modules.predict_xgboost(model_Y_test, X_train, y_train, regmode)
        theta_3_test = xgboost_modules.predict_xgboost(model_Y_train, X_test, y_test, regmode)


    theta_3_train +=  generate_noise(theta_3_train,np.mean(y_train)*(N**(-1/4))*10,1,noiseTF)
    return theta_1_train, theta_2_train, theta_3_train



# Given S, compute \omega_{S} := \prod_{i \in S} I_{vi}(Vi) / P(Vi | pa(Vi)).
def constructOmega_S(X_test, x, allCondProb, S, topoSort):
    '''
    Input
        X_test: A data for the evaluation
        x: This is v_{S}
        allCondProb: P(Vi | pa(Vi)) evaluated from X_test
        S: A set S
        topoSort: Topological order

    Output:
        omega: A dictionary key:[S1,S2,...], output \omega_{Si} := \prod_{i \in Si} I_{vi}(Vi) / P(Vi | pa(Vi)) for each Si
    '''

    eps = 1e-4
    # eps = 0
    omega = {}
    # Sorting S w.r.t topo order
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)

    # 1 vector
    prev_omega = np.repeat(1,X_test.shape[0])

    # For each v_{Si} \in v_{S}
    for idx in range(len(S_sorted)):
        # k = Si
        k = S_sorted[idx]

        ''' For each V_{S_i} \in V_{S}
        '''
        # Check Ck := { V_{Sk}+1, ..., V_{S_{k+1}} - 1} Empty
        if idx == len(S_sorted)-1:
            Ck = topoSort[topoSort.index(k) + 1:]
        else:
            k_plus_1 = S_sorted[idx + 1]
            Ck = topoSort[topoSort.index(k) + 1:topoSort.index(k_plus_1)]
        # If Ck = empty, then do nothing
        if Ck == []:
            continue

        # Identify P(Vk | pre(Vk))
        pred_k = allCondProb[k]

        # Compute I_{vk}(Vk)
        vk = x[k].values[0]
        Vk = X_test[k]
        Ivk_Vk = ((Vk == vk)*1)

        # Compute \omega_{k} = \omega_{k-1} * ( I_{vk}(Vk)/P(Vk | pa(Vk)) )
        newVal = (Ivk_Vk/(pred_k + eps))
        clip_var = 3
        newVal[newVal > clip_var ] = clip_var
        # one_sigma_newVal_lower = np.mean(newVal) - np.std(newVal)
        # one_sigma_newVal_upper = np.mean(newVal) + np.std(newVal)
        # newVal[newVal > one_sigma_newVal_upper] = one_sigma_newVal_upper  + np.random.normal(loc=0,scale=0.1,size=len(newVal))
        # newVal[newVal < one_sigma_newVal_lower] = one_sigma_newVal_lower + np.random.normal(loc=0, scale=0.1, size=len(newVal))

        newVal_prev = newVal * prev_omega
        # one_sigma_newVal_lower = np.mean(newVal_prev ) - np.std(newVal_prev )
        # one_sigma_newVal_upper = np.mean(newVal_prev ) + np.std(newVal_prev )
        # newVal_prev[newVal_prev > one_sigma_newVal_upper] = one_sigma_newVal_upper + np.random.normal(loc=0, scale=0.1,size=len(newVal))
        # newVal_prev[newVal_prev < one_sigma_newVal_lower] = one_sigma_newVal_lower + np.random.normal(loc=0, scale=0.1,size=len(newVal))
        newVal_prev[newVal_prev > clip_var] = clip_var

        omega[k] = newVal_prev

        prev_omega = omega[k]

    ''' omega[k] = \prod_{i \in S; and i \prec k}
    '''

    return omega

# Compute \theta_{k,1}, \theta_{k,2} for E[Y | do(v_S)]
def constructTheta_S(X_train, y_train, X_test, y_test, x, S, topoSort, noiseTF):
    '''
    Input
        X_train:
        y_train:
        X_test:
        y_test:
        x:
        S:

    Output
        \theta_{k,1}, \theta_{k,2} for k \in S
    '''

    N = X_test.shape[0]

    # Sort S based on the topological error.
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    S_sorted_reverse = S_sorted.copy()
    S_sorted_reverse.reverse()

    theta_1 = dict()
    theta_2 = dict()

    theta_k1_train = y_train
    theta_k1_test = y_test

    # for k in S_sorted_reverse: S_{m}, S_{m-1},...
    for idx in range(len(S_sorted_reverse)):
        k = S_sorted_reverse[idx]

        # Check Ck := { V_{Sk}+1, ..., V_{S_{k+1}} - 1} Empty
        if idx == 0:
            Ck = topoSort[topoSort.index(k) + 1:]
        else:
            k_plus_1 = S_sorted_reverse[idx-1]
            Ck = topoSort[topoSort.index(k)+1:topoSort.index(k_plus_1)]

        if Ck == []:
            continue

        # Ck = topoSort[topoSort.index(k)+1:]
        theta_1[k] = theta_k1_test

        # Vk and Predecessor of Vk; Vk will be regressed onto {Predecessor of Vk}.
        colchoice = [val for val in topoSort if topoSort.index(val) <= topoSort.index(k)]
        X_choice_train = X_train[colchoice]
        X_choice_test = X_test[colchoice]

        # Compute theta_{k,2} = E[\theta_k1 | Vk, pre(Vk)]
        regmode = xgboost_modules.checkMode(theta_k1_train)
        model = xgboost_modules.fit_xgboost(X=X_choice_train, y=theta_k1_train, regtype=regmode)
        if regmode == "binary":
            theta_k2_test = model.predict(xgboost.DMatrix(X_choice_test))
            # theta_k2_train = model.predict(xgboost.DMatrix(X_choice_train))
        else:
            theta_k2_test = xgboost_modules.predict_xgboost(model,X_choice_test,theta_k1_test,regmode)
            # theta_k2_train = xgboost_modules.predict_xgboost(model,X_choice_train,theta_k1_train,regmode)
        theta_2[k] = theta_k2_test

        # Compute theta_{k-1,1} = E[\theta_k1 | vk, pre(Vk)]
        X_choice_train = X_choice_train.copy()
        X_choice_test = X_choice_test.copy()
        X_choice_train[k] = np.repeat(x[k].values[0],X_choice_train.shape[0])
        X_choice_test[k] = np.repeat(x[k].values[0], X_choice_test.shape[0])

        if regmode == "binary":
            theta_k1_test = model.predict(xgboost.DMatrix(X_choice_test))
            theta_k1_train = model.predict(xgboost.DMatrix(X_choice_train))
        else:
            theta_k1_test = xgboost_modules.predict_xgboost(model,X_choice_test,theta_k1_test,regmode)
            theta_k1_train = xgboost_modules.predict_xgboost(model,X_choice_train,theta_k1_train,regmode)

    # This is \theta_{0,1}
    theta_1[-1] = theta_k1_test

    # Noise
    theta_1[-1] += generate_noise(theta_1[-1],np.mean(y_train)*(N**(-1/4))*10,1, noiseTF)

    # for k in theta_1.keys():
    #     theta_1[k] += N ** (-1 / 3)

    return theta_1, theta_2

################################################################################################
################################################################################################
# Double/Debiased Machine Learning
################################################################################################
################################################################################################

################################################################################################
# Case 1
################################################################################################

# Given \omega_k and \theta_{k,1}, \theta_{k,2} for all k \in S, compute the DML estimate
def computeDML_given_omega_theta(omegaDict,theta_1,theta_2):
    '''
    Input:
        omegaDict: From constructOmega_S. This is \bm{\omega}^S
        theta_1: From constructTheta_S. This is \bm{\theta^S_{1}}
        theta_2: From constructTheta_S. This is \bm{\theta^S_{2}}

    Output:
        DML estimates for E[Y | do(v_S)]
    '''
    # theta^S_{0,1}
    EIF = np.asarray(theta_1[-1].copy())

    # Sum to compute \omega_k * (\theta_{k,1} - \theta_{k,2})
    for k in omegaDict.keys():
        EIF += np.asarray(omegaDict[k]) * (np.asarray(theta_1[k]) - np.asarray(theta_2[k]))
    return np.mean(EIF)

# Compute the DML estimator for v_{S}; i.e., Compute E[Y | do(V_{S} = v_{S})] using DML
def computeDML_S(X_list,y_list, allCondProb_list, x, S, topoSort, DMLDict, noiseTF):
    '''
    Input
        X_list: [X_{train}, X_{test}]
        y_list: [y_{train}, y_{test}]
        allCondProb_list: [allCondProb_eval_train, allCondProb_eval_test]
        x: Fixed v_{S}
        S: a subset S
        topoSort:
        DMLDict: key: S, value: DML estimates

    Output
        DMLDict
    '''

    # If S = 0 or S = [n], do nothing
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        DMLDict[keygen(S)] = 1
        return DMLDict

    # Sort S w.r.t. topoOrder
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)

    # If S is in DMLDict, then just return DMLDict
    if keygen(S_sorted) in DMLDict.keys():
        return DMLDict

    result = 0
    for k in [0,1]:
        # Compute \bm{\omega^S};
        ## Note allCondProb is trained with X_list[1-k] and evaluated with X_list[k]
        omegaDict = constructOmega_S(X_test =X_list[k], x=x, allCondProb=allCondProb_list[k], S=S, topoSort=topoSort)

        # Compute \bm{\theta}^S_{0}, \bm{\theta}^S_{1}
        theta_1, theta_2 = constructTheta_S(X_train = X_list[1-k], y_train = y_list[1-k], X_test=X_list[k], y_test = y_list[k], x=x, S=S, topoSort=topoSort, noiseTF = noiseTF)
        result += computeDML_given_omega_theta(omegaDict, theta_1, theta_2)

    result /= 2
    DMLDict[keygen(S_sorted)] = result
    return DMLDict

################################################################################################
# Case 2
################################################################################################
def computeDML_S_Case2(X_list,y_list, model_Y_list, x, S, topoSort, DMLDict, noiseTF):
    '''
    Input
        X_list: [X_{train}, X_{test}]
        y_list: [y_{train}, y_{test}]
        allCondProb_list: [allCondProb_eval_train, allCondProb_eval_test]
        x: Fixed v_{S}
        S: a subset S
        topoSort:
        DMLDict: key: S, value: DML estimates

    Output
        DMLDict
    '''

    # If S = 0 or S = [n], do nothing
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        DMLDict[keygen(S)] = 1
        return DMLDict

    # Sort S w.r.t. topoOrder
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)

    # If S is in DMLDict, then just return DMLDict
    if keygen(S_sorted) in DMLDict.keys():
        return DMLDict

    result = 0
    for k in [0,1]:
        # X_train = X_list[k];
        X_train = X_list[k]
        X_test = X_list[1-k]
        y_train = y_list[k]
        y_test = y_list[1 - k]
        model_Y_train = model_Y_list[k] # model trained using train dataset
        model_Y_test = model_Y_list[1-k] # model trained using test dataset

        # Train Cond Probl
        allCondModel_S = train_condProb_case2(X_list[k], topoSort, S_sorted)
        evalCondModel_S = estimate_condProb_case2(X_list[1-k],allCondModel_S, S, topoSort, noiseTF)

        # compute ometaDict
        omega = constructOmega_S(X_list[1-k], x, evalCondModel_S, S_sorted, topoSort)
        omega = omega[S_sorted[-1]]

        # compute theta
        theta_1,theta_2,theta_3 = constructTheta_S_case2(X_train, y_train, X_test, y_test, model_Y_train, model_Y_test, x, S, topoSort, noiseTF)

        # Compute \bm{\theta}^S_{0}, \bm{\theta}^S_{1}
        EIF = np.array(theta_3) + (np.asarray(omega) * (np.array(theta_1) - np.array(theta_2)))
        result += np.mean(EIF)
    result /= 2
    DMLDict[keygen(S_sorted)] = result
    return DMLDict

################################################################################################
################################################################################################
# Inverse Probability Weighting (IPW)
################################################################################################
################################################################################################
def computeDML_given_omega_IPW(y,omegaDict,S,topoSort):
    '''
    Input
        :omegaDict: constructOmega_S.

    Output
        IPW estimate
    '''
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    return np.mean( omegaDict[S_sorted[-1]] * y )


def computeIPW_S(X_list,y_list, allCondProb_list, x, S, topoSort, IPWDict):
    # If S = emptyset or S is a all feature set,
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        IPWDict[keygen(S)] = 1
        return IPWDict

    # Preprocessing the S
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    # If S exists in the dictionary
    if keygen(S_sorted) in IPWDict.keys():
        return IPWDict

    result = 0
    for k in [0, 1]:
        omegaDict = constructOmega_S(X_test=X_list[k], x=x, allCondProb=allCondProb_list[k], S=S, topoSort=topoSort)
        result += computeDML_given_omega_IPW(y_list[k], omegaDict,S,topoSort)

    result /= 2
    IPWDict[keygen(S_sorted)] = result
    return IPWDict

################################################################################################
# Case 2
################################################################################################
def computeIPW_S_case2(X_list,y_list, x, S, topoSort, IPWDict, noiseTF):
    # If S = emptyset or S is a all feature set,
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        IPWDict[keygen(S)] = 1
        return IPWDict

    # Preprocessing the S
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    # If S exists in the dictionary
    if keygen(S_sorted) in IPWDict.keys():
        return IPWDict

    result = 0
    for k in [0, 1]:
        # Train Cond Probl
        allCondModel_S = train_condProb_case2(X_list[k], topoSort, S_sorted)
        evalCondModel_S = estimate_condProb_case2(X_list[1 - k], allCondModel_S, S, topoSort,noiseTF)

        # compute ometaDict
        omega = constructOmega_S(X_list[1 - k], x, evalCondModel_S, S_sorted, topoSort)
        omega = omega[S_sorted[-1]]
        result += np.mean( y_list[1-k] * omega )

    result /= 2
    IPWDict[keygen(S_sorted)] = result
    return IPWDict





################################################################################################
# REG Shapley
################################################################################################
def computeDML_given_theta1_PI(theta_1):
    '''
    Input
        :omegaDict: constructOmega_S.

    Output
        PI estimate
    '''
    return np.mean( theta_1[-1] )

def computePI_S(X_list,y_list, x, S, topoSort, PIDict, noiseTF):
    # If S = emptyset or S is a all feature set,
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        PIDict[keygen(S)] = 1
        return PIDict

    # Preprocessing the S
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    # If S exists in the dictionary
    if keygen(S_sorted) in PIDict.keys():
        return PIDict

    result = 0
    for k in [0, 1]:
        theta_1, theta_2 = constructTheta_S(X_train = X_list[k], y_train = y_list[k], X_test = X_list[1-k], y_test = y_list[1-k], x=x, S=S, topoSort=topoSort, noiseTF = noiseTF)
        result += computeDML_given_theta1_PI(theta_1)

    result /= 2
    PIDict[keygen(S_sorted)] = result
    return PIDict

def computePI_S_case2(X_list,y_list, model_Y_list, x, S, topoSort, PIDict, noiseTF):
    # If S = emptyset or S is a all feature set,
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        PIDict[keygen(S)] = 1
        return PIDict

    # Preprocessing the S
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    # If S exists in the dictionary
    if keygen(S_sorted) in PIDict.keys():
        return PIDict

    result = 0
    for k in [0, 1]:
        theta_1, theta_2, theta_3 = constructTheta_S_case2(X_train = X_list[k], y_train = y_list[k], X_test = X_list[1-k], y_test = y_list[1-k], model_Y_train = model_Y_list[k], model_Y_test = model_Y_list[1-k],x=x,S=S,topoSort=topoSort, noiseTF = noiseTF)
        result += np.mean(theta_3)

    result /= 2
    PIDict[keygen(S_sorted)] = result
    return PIDict



################################################################################################
# Sampling based (Heskes)
################################################################################################
def computeSampling_S(X_list,y_list, allCondModel_list, model_Y_list, x, S, topoSort, parentDict, SampleDict, noiseTF):
    # If S = emptyset or S is a all feature set,
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        SampleDict[keygen(S)] = 1
        return SampleDict

    N = X_list[0].shape[0]

    # Preprocessing the S
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    # If S exists in the dictionary
    if keygen(S_sorted) in SampleDict.keys():
        return SampleDict

    # Establish {vi}_{i \in S}
    newX_list = X_list.copy()
    numSample = 10
    sampled_data = dict()
    result = 0

    for p in [0,1]:
        myX = newX_list[p]
        # Call the model { P(Vi | pre(Vi)) }
        allCondModel = allCondModel_list[p]
        model_Y = model_Y_list[p]

        for idx in range(len(topoSort)-1):
            # Set Vk, vk
            k = topoSort[idx]
            Vk = myX[k]
            uniqueVk = sorted(pd.unique(Vk))
            vk = x[k].values[0]

            # If k \in S, then P_{\pi(k)}(Vk | pre(Vk)) = I_{vk}(Vk)
            # Therefore, the sample is just vk
            if k in S_sorted:
                # Make a probability vector P_{\pi(k)}(Vk | pre(Vk))
                problist = [0] * len(uniqueVk)
                problist[uniqueVk.index(vk)] = 1

                # Sample Vk
                sampled_vk = np.random.choice(len(uniqueVk), numSample, p=problist)
                # Append the sampled Vk into the dictionary sampled_data
                sampled_data[k] = sampled_vk

            # If k \not \in S, then P_{\pi(k)}(Vk | pre(Vk)) = P(Vk | pre(Vk))
            else:
                # Extract the model for P(Vk | pre(Vk))
                model_k = allCondModel[k]

                ## if pre(Vk) == [], then easy to sample
                if idx == 0:
                    model_k = model_k / np.sum(model_k)
                    # Sample Vk from the prob vector for P(vk)
                    sampled_vk = np.random.choice(len(model_k), numSample, p=model_k)

                    # Append
                    sampled_data[k] = sampled_vk

                ## otherwise, sample from P(Vk | pre(Vk))
                else:
                    ### Construct pre(Vk)
                    sampled_data_df = pd.DataFrame.from_dict(sampled_data)
                    preVk = sampled_data_df[topoSort[:idx]]

                    ### Check Vk is binary or not
                    regmode = xgboost_modules.checkMode(Vk)
                    if regmode == 'binary':
                        # P(Vk = 1 | pre(Vk))
                        problist = model_k.predict(xgboost.DMatrix(preVk))
                        # Add noise
                        # problist += generate_noise(problist,10*(N**(-1/4)),1)
                        # norm_fun = lambda myp: abs(myp) / np.sum(abs(myp))
                        # problist = np.apply_along_axis(norm_fun, 1, problist)

                        sample_fun = lambda myp: np.random.choice(2, 1, p=[1-myp,myp])[0]
                        sampled_data[k] = np.array(list(map(sample_fun,problist)))
                    else:
                        problist = model_k.predict(xgboost.DMatrix(preVk))
                        problist += generate_noise(problist,10*(N**(-1/4)),1, noiseTF)
                        norm_fun = lambda myp: abs(myp) / np.sum(abs(myp))
                        problist = np.apply_along_axis(norm_fun,1,problist)

                        sample_fun = lambda myp: np.random.choice(problist.shape[1], 1, p=myp)[0]
                        sampled_data[k] = np.apply_along_axis(sample_fun,1,problist)

        sampled_data = pd.DataFrame.from_dict(sampled_data)

        ''' If Y is continuous 
        '''
        expec_Y_list = model_Y.predict(xgboost.DMatrix(sampled_data))
        # NOISE
        expec_Y_list += generate_noise(expec_Y_list, (N ** (-1 / 4)) * 10*np.mean(y_list[p]),1, noiseTF)

        sigma_Y = 1
        sampled_Y = sp.stats.norm.rvs(loc=expec_Y_list, scale=sigma_Y)
        result += np.mean(sampled_Y)

        # ''' If Y is binary
        # '''
        # prob_Y_list = model_Y.predict(xgboost.DMatrix(sampled_data))
        # # NOISE
        # prob_Y_list += N ** (-1 / 3)
        # prob_Y_list[prob_Y_list <= 0] = 1e-4
        # prob_Y_list[prob_Y_list >= 1] = 1-1e-4
        # sample_Y_fun = lambda myp: np.random.choice(2, 1, p=[1 - myp, myp])[0]
        # result += np.mean(pd.DataFrame(prob_Y_list).applymap(sample_Y_fun)[0])
    result /= 2
    SampleDict[keygen(S_sorted)] = result
    return SampleDict


if __name__ == '__main__':
    n = 1000
    seednum = 123
    targetidx = 0

    X, topoSort, dictName, parentDict = data_generator_2.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X_test.iloc[[targetidx]]
    y = y_test.iloc[[targetidx]].values[0]

    X_list = [X_train, X_test]
    y_list = [y_train, y_test]

    S = [5,8,2]
    model_Y_train = train_Y(X_train,y_train)
    model_Y_test = train_Y(X_test,y_test)
    model_Y_list = [model_Y_train, model_Y_test]
    DMLDict = {}
    DMLDict = computeDML_S_Case2(X_list, y_list, model_Y_list, x, S, topoSort, DMLDict)

    IPWDict = {}
    IPWDIct = computeIPW_S_case2(X_list, y_list, x, S, topoSort, IPWDict)

    PIDict = {}
    PIDict = computePI_S_case2(X_list, y_list, model_Y_list, x, S, topoSort, IPWDict)
