# This cell defines the functions we use to generate the data in our scenario

import numpy as np
import pandas as pd
import scipy.stats
import sklearn
import xgboost
from datagen_modules import discretizer
import shap
import matplotlib.pyplot as plt
import data_generator_1
import datagen_modules

class FixableDataFrame(pd.DataFrame):
    """ Helper class for manipulating generative models.
    """
    def __init__(self, *args, fixed={}, **kwargs):
        self.__dict__["__fixed_var_dictionary"] = fixed
        super(FixableDataFrame, self).__init__(*args, **kwargs)
    def __setitem__(self, key, value):
        out = super(FixableDataFrame, self).__setitem__(key, value)
        if isinstance(key, str) and key in self.__dict__["__fixed_var_dictionary"]:
            out = super(FixableDataFrame, self).__setitem__(key, self.__dict__["__fixed_var_dictionary"][key])
        return out

# generate the data
def generator(n, fixed={}, seed=0):
    """ The generative model for our subscriber retention example.
    """
    if seed is not None:
        np.random.seed(seed)
    X = FixableDataFrame(fixed=fixed)

    numbins = 5

    # the number of sales calls made to this customer
    X["Sales calls"] = discretizer(np.random.uniform(0, 4, size=(n,)).round(), numbins=numbins)

    # the number of sales calls made to this customer
    X["Interactions"] = discretizer(X["Sales calls"] + np.random.poisson(0.2, size=(n,)), numbins)

    # the health of the regional economy this customer is a part of
    X["Economy"] = discretizer(np.random.uniform(0, 1, size=(n,)), numbins)

    # the time since the last product upgrade when this customer came up for renewal
    X["Last upgrade"] = discretizer(np.random.uniform(0, 20, size=(n,)), numbins)

    # how much the user perceives that they need the product
    X["Product need"] = discretizer(X["Sales calls"] * 0.1 + np.random.normal(0, 1, size=(n,)), numbins)

    # the fractional discount offered to this customer upon renewal
    X["Discount"] = (1-scipy.special.expit(X["Product need"])) * 0.5 + 0.5 * np.random.uniform(0, 1, size=(n,)) / 2


    # What percent of the days in the last period was the user actively using the product
    X["Monthly usage"] = discretizer(scipy.special.expit(X["Product need"] * 0.3 + np.random.normal(0, 1, size=(n,))), numbins)

    # how much ad money we spent per user targeted at this user (or a group this user is in)
    X["Ad spend"] = discretizer(X["Monthly usage"] * np.random.uniform(0.99, 0.9, size=(n,)) + (X["Last upgrade"] < 1) + (X["Last upgrade"] < 2), numbins)

    # how many bugs did this user encounter in the since their last renewal
    X["Bugs faced"] = discretizer(np.array([np.random.poisson(v*2) for v in X["Monthly usage"]]), numbins)

    # how many bugs did the user report?
    X["Bugs reported"] = (X["Bugs faced"] * scipy.special.expit(X["Product need"])).round()

    # did the user renew?
    X["Did renew"] = scipy.special.expit(7 * (
          0.18 * X["Product need"] \
        + 0.08 * X["Monthly usage"] \
        + 0.1 * X["Economy"] \
        + 0.05 * X["Discount"] \
        + 0.05 * np.random.normal(0, 1, size=(n,)) \
        + 0.05 * (1 - X['Bugs faced'] / 20) \
        + 0.005 * X["Sales calls"] \
        + 0.015 * X["Interactions"] \
        + 0.1 / (X["Last upgrade"]/4 + 0.25)
        + X["Ad spend"] * 0.05 - 0.45
    ))

    # in real life we would make a random draw to get either 0 or 1 for if the
    # customer did or did not renew. but here we leave the label as the probability
    # so that we can get less noise in our plots. Uncomment this line to get
    # noiser causal effect lines but the same basic results
    X["Did renew"] = scipy.stats.bernoulli.rvs(X["Did renew"])

    return X


def user_retention_dataset():
    """ The observed data for model training.
    """
    n = 10000
    X_full = generator(n)
    y = X_full["Did renew"]
    X = X_full.drop(["Did renew", "Product need", "Bugs faced"], axis=1)
    return X, y

def fit_xgboost(X, y):
    """ Train an XGBoost model with early stopping.
    """
    X_train,X_test,y_train,y_test = sklearn.model_selection.train_test_split(X, y)
    dtrain = xgboost.DMatrix(X_train, label=y_train)
    dtest = xgboost.DMatrix(X_test, label=y_test)
    model = xgboost.train(
        {"eta": 0.001, "subsample": 0.5, "max_depth": 2, "objective": "reg:squarederror"}, dtrain, num_boost_round=200000,
        evals=((dtest, "test"),), early_stopping_rounds=20, verbose_eval=False
    )
    # model = xgboost.train(
    #     { "eta": 0.001, "subsample": 0.5, "max_depth": 2, "objective": "reg:logistic"}, dtrain, num_boost_round=200000,
    #     evals=((dtest, "test"),), early_stopping_rounds=20, verbose_eval=False
    # )
    return model

def Run_Cond(n,seednum,numiter_per_sim,num_sim,targetidx,num_process):
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    Y = X[X.columns[-1]]
    X = X[X.columns[:-1]].copy()
    model = fit_xgboost(X,Y)

    # X, y = user_retention_dataset()
    # model = fit_xgboost(X, y)

    explainer = shap.Explainer(model)
    shap_values = explainer(X.iloc[[targetidx]])
    # shap_values_all = explainer(X)
    # shap.plots.scatter(shap_values_all[:,"Discount"])
    condDict = dict()

    for idx in range(len(X.columns)):
        # print(dictName[idx], ":", np.round(shap_values.values[0, idx], 4))
        condDict[idx] = np.round(shap_values.values[0, idx], 4)

    return condDict

if __name__ == '__main__':
    n = 10000
    seednum = 123
    targetidx = 10
    num_process = 5

    numiter_per_sim = 10
    num_sim = 10

    Run_Cond(n, seednum, numiter_per_sim, num_sim, targetidx, num_process)

# print(shap_values.values[targetidx,4])
#
# clust = shap.utils.hclust(X, y, linkage="complete")
# shap.plots.bar(shap_values, clustering=clust, clustering_cutoff=1)
#
# # import datagen_modules
# # X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
# # x = X_test.iloc[[targetidx]]
# # y = y_test.iloc[[targetidx]].values[0]
#
# shap.plots.scatter(shap_values[:,"Discount"])

